(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, ENAC/CENA          *)
(*                                                                     *)
(***********************************************************************)
(* $Id: fcl_alldiff.ml,v 1.9 2000/12/22 14:55:43 barnier Exp $ *)

open Fcl_var
open Fcl_arith

let such_that f t =
  let n = Array.length t in
  let rec st i =
    if i < n
    then if f t.(i) then i::st (i+1) else st (i+1)
    else [] in
  st 0;;

class diff_from_others i (vars : Fd.t array) = object
  inherit Fcl_cstr.t Fcl_cstr.immediate

  method name = "diff_from_others"
  method delay x = delay [Attr.on_subst] vars.(i) x
  method check = raise Fcl_cstr.DontKnow

  method update =
    match Fd.value vars.(i) with
      Unk _ -> ()
    | Val x ->
	Fcl_stak.set solved true;
	for j = 0 to Array.length vars - 1 do
	  if i <> j
	  then match Fd.value vars.(j) with
	    Unk a -> Fd.refine vars.(j) (Fcl_domain.remove x (Attr.dom a))
	  | Val y -> if x = y then Fcl_stak.fail "diff_from_others"
	done
end

module Int = struct
  type t = int
  let equal (x : int) y = x = y
  let hash (x : int) = Hashtbl.hash x
end

module IntHashtbl = Hashtbl.Make(Int)
    

class diff (vars : Fd.t array) on_event = 
  let n = Array.length vars
      (* Une valeur en dehors des valeurs possibles *)
  and valout = Array.fold_right ~f:(fun v acc -> min acc (Fd.min v)) vars ~init:max_int - 1 in
  let matchingl = Array.create n valout in (* var -> value : not backtrackable *)
  let matchingr = IntHashtbl.create n in(* value -> var : not backtrackable *)

  object
  inherit Fcl_cstr.t Fcl_cstr.later

  method name = "Permut.diff"
  method delay x = Array.iter ~f:(fun v -> delay [on_event] v x) vars
  method check = raise Fcl_cstr.DontKnow

  method update =
    let getmatchingr i = try IntHashtbl.find matchingr i with Not_found -> -1 in
      (* matching update (values have been removed from the domains) *)
    let lefts = ref [] in
    for i = n-1 downto 0 do (* downto to get lefts ordered *)
      if matchingl.(i) = valout
	then lefts := i :: !lefts
      else match Fd.value vars.(i) with
	Unk a ->
	  if not (Fcl_domain.member matchingl.(i) (Attr.dom a)) then begin
	    IntHashtbl.remove matchingr matchingl.(i);
	    matchingl.(i) <- valout;
	      lefts := i :: !lefts
	  end
      | Val x ->
	  if x <> matchingl.(i)
	  then begin
	    let y = getmatchingr x in
	    if y <> -1 (* Value x was already used by y *)
	    then begin
	      matchingl.(y) <- valout;
	      IntHashtbl.remove matchingr x;
	      lefts := y :: !lefts
	    end;
	    if matchingl.(i) <> valout then (*=valout the first time*)
	      IntHashtbl.remove matchingr matchingl.(i);
	    matchingl.(i) <- x;
	    assert(not (IntHashtbl.mem matchingr x));
	    IntHashtbl.add matchingr ~key:x ~data:i
	  end
    done;
    if !lefts <> [] then begin
      let apath = IntHashtbl.create n in
      let getapath i = try IntHashtbl.find apath i with Not_found -> -1 in
      let reset_apath () = IntHashtbl.clear apath in
      
      let depth_first rights =
	let rec reverse right left =
	  let r = matchingl.(left) in
	  matchingl.(left) <- right;
	  if r <> valout
	  then begin
	    let a = IntHashtbl.find apath r in
	    assert(IntHashtbl.mem matchingr r);
	    IntHashtbl.remove matchingr r;
	    IntHashtbl.add matchingr ~key:r ~data:a;
	    IntHashtbl.remove apath r;
	    reverse r a
	  end
	and check r =
	  let ar = getapath r in
	  ar <> -1 &
	  (matchingl.(ar) = valout or check (matchingl.(ar))) in
	List.iter
	  ~f:(fun r ->
	    if getmatchingr r = -1 & check r then begin
	      let a = IntHashtbl.find apath r in
	      assert(not (IntHashtbl.mem matchingr r));
	      IntHashtbl.add matchingr ~key:r ~data:a;
	      IntHashtbl.remove apath r;
	      reverse r a
	    end)
	  rights
      in
      
      let rec breadth_first lefts =
	lefts <> [] && begin
	  let ending = ref false
	  and rights = ref []
	  and new_lefts = ref [] in
	  List.iter
	    ~f:(fun l ->
	      try
        	Fd.iter vars.(l)
		  (fun r ->
		    if getapath r = -1 then begin
		      assert(not (IntHashtbl.mem apath r));
		      IntHashtbl.add apath ~key:r ~data:l;
		      rights := r :: !rights;
		      let mr = getmatchingr r in
		      if mr = -1
		      then raise Exit
		      else new_lefts := mr :: !new_lefts
		    end)
	      with
		Exit -> ending := true)
	    lefts;
	  if !ending
	  then begin
	    depth_first !rights;
	    let lefts = such_that (fun xi -> xi = valout) matchingl in
	    lefts = [] or (reset_apath (); breadth_first lefts)
	  end else breadth_first !new_lefts
	end in
      if not (breadth_first !lefts) then Fcl_stak.fail "permut"
    end
end

type algo = Lazy | Bin_matching of Attr.event

let cstr ?(algo = Lazy) vars =
  let n = Array.length vars in
  if n <= 1 then Fcl_cstr.one else
    let sum =
      ref (match algo with
	       Bin_matching on_event ->
		 fd2e (Fcl_reify.boolean ~delay_on_negation:false
			 (new diff vars on_event))
	     | Lazy -> i2e 1) in
    for i = 0 to n - 1 do
      sum := !sum +~
      fd2e (Fcl_reify.boolean ~delay_on_negation:false
	      (new diff_from_others i vars))
    done;
    !sum =~ i2e (Array.length vars + 1);;
  
   
