(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, ENAC/CENA          *)
(*                                                                     *)
(***********************************************************************)
(* $Id: fcl_cstr.mli,v 1.20 2001/01/29 11:00:33 brisset Exp $ *)

(* Module [Cstr]: posting constraints and building new ones *)

(*** Overview *)
(*q This module defines the type [t] of constraints and functions to create
  and post constraints: mainly a [create] function which allows to build new
  constraints from scratch (this function is not needed when using
  standard FaCiLe predefined constraints) and the mostly useful [post]
  function which must be called to effectively add a constraint to the
  constraints store.
*)

(*** Basic *)
exception DontKnow
(* Exception raised by the [check] function of a constraint (of type
   [t]) when it is not known whether the constraint is satisfied or
   violated. *)
type priority
(* Waking priority. *)
val immediate : priority
(* As soon as possible, for quick updates. *)
(*{*) (* e.g. for deamons *) (*}*)
val normal : priority
(* Standard priority. *)
val later : priority
(* For time consuming constraints (e.g. [Gcc,cstr], [Alldiff.cstr], ...). *)

(*{*)
(* The mother class of all constraints. *)
class virtual t : priority -> object
  val solved : bool Fcl_stak.ref
(* To be set to [True] as soon as the constraint cannot be violated. *)
  method id : int
(* A unique integer identifying the constraint object. *)
  method woken : bool Fcl_stak.ref
(* _Undocumented_
   Used by [wake] and [wake_all] to check wether the constraint object is
   already active during processing the constraints queue. *)
  method name : string
(* Describing name of the constraint. *)
  method is_solved : bool
(* _Undocumented_
   Used by [wake] during constraints queue processing and by suspension lists
   of variables (of type [Var.cstrd]) by their [delay] method. *)
  method virtual update : unit
(* [update] compute the domain reductions of the constraint, should raise
   [Stak.Fail] whenever a failure occurs and position [solved] to
   [true] when the constraint is solved ([Stak.set solved true]). *)
  method init : unit
(* Used to initialize inner data structures of the constraint needed by the
   update method. First method called by [post]. Default : [()]. *)
  method check : bool
(* Used by the [update] method of reified constraints to check if the
   constraint is satisfied or violated. Should raise [DontKnow] when it is
   not known wether the constraint is satisfied or violated.
   Default : [Failure] exception raised. *)
  method not : t
(* Used by the [update] method of reified constraint to post the negation of
   the constraint when it is violated.
   Default: [Failure] exception raised. *)
  method virtual delay : t -> unit
(* [delay] schedule the awakening of the constraint object [c],
   i.e. the execution of its [method] update. If [update] should be
   called (because it may reduce variables domains) when one of the
   events of the list [es] occured on variable [v], then
   [Var.delay es v c] should be called within the body of method
   [delay]. *)
(* _Undocumented_
   Usually called with itself as parameter, i.e. [c#delay c] (except
   within the reification module [Reify]), to schedule the awakening of the
   constraint object [c]. i*)
  method fprint : out_channel -> unit
(* [fprint chan] print the constraint object on channel [chan]. *)
  method priority : priority
(* Return the priority of the constraint object. *)
end
(*}*)


(*{*)
(* Hack to document the abstract type t (appearing in facile.mli) *)
(*}*)
(*q
[
type t]
The type of constraints
*)

val id : t -> int
(* [id c] returns a unique integer identifying the constraint [c]. *)

val name : t -> string
(* [name c] returns the name of the constraint [c]. *)

val fprint : out_channel -> t -> unit
(* [fprint chan c] prints the constraint [c] on channel [chan]. Calls
   the [fprint] function passed to [create]. *)

val create : ?name:string -> ?fprint:(out_channel -> unit) -> ?priority:priority -> ?init:(unit -> unit) -> ?check:(unit -> bool) -> ?not:(unit -> t) -> (unit -> bool) -> (t -> unit) -> t
(* [create ?name ?fprint ?priority ?init ?check ?not update delay]
   builds a new constraint.

-  [name] is a describing string name of the constraint. Default value
   is ["anonymous"].

-  [fprint] (pretty-)printing of the constraint on an output channel
   taken as its only argument. Default value is to print the [name] string.

-  [priority] is either [immediate], [normal] or [later]. Time costly
   constraints should be waken after quick ones. Default value is [normal].

-  [init] is useful to perform initialization of auxiliary data
   structures needed and maintained by the [update] function.
   [init ()] is called as soon as the constraint is posted. Default
   value is [fun () -> ()].

-  [check] must be specified if the constraint is reifiable (as well
   as [not]). When the constraint is reified, [check ()] is called to
   verify whether the
   constraint is satisfied or violated, i.e. the constraint itself or
   its negation is entailed by the constraint store, and should return
   [true] if the constraint is satisfied, [false] if it is violated
   and raise [DontKnow] when it is not known. [check] must not change
   the domains of the variables involved in the constraint. Default :
   [Failure] exception raised.

-  [not] must be specified if the constraint is reifiable (as well
   as [check]). [not ()] should return a constraint which is the
   negation of the constraint being defined. When the constraint is
   reified, it is called to post the
   negation of the constraint whenever [check ()] return [false],
   i.e. it is entailed by the constraint store. Default : [Failure]
   exception raised.

-  [update] computes the domain reductions of the constraint. [update ()]
   should return [true] when the constraint become solved, [false] if
   it is not yet entailed by the constraint store and raise 
   [Stak.Fail] whenever a failure occurs. [update] is a mandatory
   argument.

-  [delay] schedules the awakening of the constraint [ct] (which is
   taken as its only argument), i.e. the execution of its [update]
   function. If [update ()] should be called (because it may reduce
   variables domains) when one of the events contained in the events
   list [es] occured on variable [v], then [Var.delay es v ct] should
   be called within the body of the [delay] function. [delay] is a mandatory
   argument.
*)

(*{*)
val reset_queue : unit -> unit
(* _Undocumented_
   [reset_queue ()] reset the constraint queue. *)
val wake : t -> unit
(* _Undocumented_
   [wake c] push the constraint [c] on the constraint awakening priority
   queue. *)
val wake_all : unit -> unit
(* _Undocumented_
   [wake_all ()] wake all constraints respecting priority order. *)
(*}*)
val post : t -> unit
(* [post c] posts the constraint [c] to the constraint store. *)
(*{*)
val init : t -> unit
(* _Undocumented_
   [init c] post the constraint deamon [c] (no wake and no add call). *)
(*}*)

val one : t
(* The constraint which succeeds immediately. *)
val zero : t
(* The constraint which fails immediately. *)

val active_store : unit -> t list
(* [active_store ()] returns the list of all active constraints, i.e. whose
   [update] functions have returned [false]. *)
