(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, ENAC/CENA          *)
(*                                                                     *)
(***********************************************************************)
(* $Id: fcl_domain.mli,v 1.22 2001/02/06 14:17:21 barnier Exp $ *)

(* Module [Domain]: domain operations *)

(* This module provides all necessary functions (and more) to create
   and handle domains, which are needed to create variables and perform
   propagation (i.e. domain reduction or filtering). *)
   

type t
(* Type of finite domains of integers (functional : no in place
   modifications, domains can be shared). Standard equality and
   comparison can be used on type domain. *)

(*** Building new domains *)
val empty : t
(* The empty domain. *)
val create : int list -> t
(* [create l] builds a new domain containing the values of [l]. Removes
   duplicates and sorts values. Returns [empty] if [l] is
   empty. *)
val interval : int -> int -> t
  (* [interval inf sup] returns the domain of all integer in the closed
     interval [inf]..[sup]. Raise [Invalid_argument] if [inf > sup]. *)
val int : t
  (* The largest representable domain. *)
val boolean : t
  (* The domain containing [0] and [1]. *)

(*** Access functions *)
val is_empty : t -> bool
(* [is_empty d] tests if the domain [d] is empty. *)
val size : t -> int
(* [size d] returns the number of integers in [d]. *)
val min : t -> int
(* [min d] returns the lower bound of [d]. If [d] is empty, the
   behaviour is unspecified (incorrect return value or exception
   raised). *)
val max : t -> int
(* [max d] returns the upper bound of [d]. If [d] is empty, the
   behaviour is unspecified (incorrect return value or exception
   raised). *)
val iter : (int -> unit) -> t -> unit
  (* [iter f d] applies function [f] in turn to all element of [d] in
  increasing order *)
val member : int -> t -> bool
(* [member n d] tests if [n] belongs to [d]. *)
val values : t -> int list
  (* [value d] returns the list of values of the domain [d] *)
val fprint : out_channel -> t -> unit
(* [fprint chan d] prints [d] on channel [chan]. *)
val sprint : t -> string
(* [sprint d] returns a string representation of [d]. *)
val included : t -> t -> bool
  (* [included d1 d2] tests if domain [d1] is included in domain [d2]. *)
val first_geq_value : t -> int -> int
  (* [first_geq_value dom val] returns the smallest value in [dom] greater
     or equal to [val]. Raises [Not_found] if [max dom < val]. *)
val first_leq_value : t -> int -> int
  (* [first_leq_value dom val] returns the greatest value in [dom] smaller
     or equal than [val]. Raises [Not_found] if [min dom > val]. *)
val choose : (int -> int -> bool) -> t -> int
  (* [choose ord d] returns the mininum value of [d] for order [ord].
     Raises [Not_found] if [d] is empty. *)

(*** Operations *)
val remove : int -> t -> t
(* [remove n d] returns [d]-{[n]}. Returns [d] if [n] is not in [d] *)
val add : int -> t -> t
(* [add n d] returns [d]+{[n]}. *)
val remove_up : int -> t -> t
  (* [remove_up n d] returns [d]-{[n]+1..[max_int]}, i.e. removes values
     stricly greater than [n]. *)
val remove_low : int -> t -> t
  (* [remove_low n d] returns [d]-{[min_int]..[n]-1}, i.e. removes values
     stricly less than [n]. *)
val remove_closed_inter : int -> int -> t -> t
  (* [remove_closed_inter inf sup d] returns [d]-{[inf]..[sup]}, i.e. removes
     values greater or equal to [inf] and less or equal to [sup] in [d].
     Returns [d] if [inf > sup]. *)
val intersection : t -> t -> t
(* Intersection over domains. *)
val union : t -> t -> t
(* Union over domains. *)
val difference : t -> t -> t
    (* [difference big small] returns [big]-[small]. [small] must be included
       in [big], otherwise the behaviour is unspecified (incorrect return
       value or exception raised). *)
val remove_min : t -> t
(* [remove_min d] returns [d] without its lower bound. Raises
   [Invalid_argument] if [d] is empty. *)
val remove_max : t -> t
(* [remove_min d] returns [d] without its upper bound. Raises
   [Invalid_argument] if [d] is empty. *)
val minus : t -> t
  (* [minus d] returns the domain of opposite values of [d]. *)
val plus : t -> int -> t
  (* [plus d n] translates a domain by [n]. *)

