(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, ENAC/CENA          *)
(*                                                                     *)
(***********************************************************************)
(* $Id: fcl_fdArray.ml,v 1.9 2001/01/29 14:10:00 brisset Exp $ *)
open Fcl_var
open Fcl_arith

class min_array xs x =
  let n = Array.length xs in
 object
  inherit Fcl_cstr.t Fcl_cstr.normal
  method name = "min_array"
  method delay c =
    Array.iter ~f:(fun x -> delay [Attr.on_min;Attr.on_max] x c) xs;
    delay [Attr.on_min;Attr.on_max] x c
  method update =
    (* let's try to decide which one is the smallest *)
    let smallest = ref 0 and min_smallest = ref (Fd.min xs.(0)) in
    for i = 1 to n - 1 do
      let min_i = Fd.min xs.(i) in
      if min_i < !min_smallest then begin
	smallest := i; min_smallest := min_i
      end
    done;
    let max_smallest = Fd.max xs.(!smallest) in
    try
      for i = 0 to n - 1 do
	if i <> !smallest && Fd.min xs.(i) < max_smallest then
	  raise Exit
      done;
      Fcl_stak.set solved true;
      Fcl_cstr.post (fd2e x =~ fd2e xs.(!smallest))
    with
      Exit ->
	(* All the xs are greater than the min of min : *)
      	let minx = Fd.min x in
      	let x_ge_min xi =
	  match Fd.value xi with
	    Val v -> if not (minx <= v) then Fcl_stak.fail "Arith.min"
	  | Unk a ->
	      if not (Fd.min x <= Attr.min a) then
	      	Fd.refine xi (Fcl_domain.remove_low minx (Attr.dom a)) in
	Array.iter ~f:x_ge_min xs;

	(* smallest min of xs <= x <= smallest max of xs : *)
      	let smallest_max =
	  Array.fold_left ~f:(fun r xi -> min (Fd.max xi) r) xs ~init:max_int in
	match Fd.value x with
	  Val x ->
	    if not (!min_smallest <= x && x <= smallest_max) then
	      Fcl_stak.fail "Arith.min"
      	| Unk a ->
	    let d = Fcl_domain.intersection (Attr.dom a) (Fcl_domain.interval !min_smallest smallest_max) in
	    Fd.refine x d
end

let min_cstr xs x =
  if Array.length xs = 0 then invalid_arg "FdArray.min_cstr";
  new min_array xs x;;

let min xs =
  let x = Fd.create Fcl_domain.int in
  let c = min_cstr xs x in
  Fcl_cstr.post c;
  x;;

let minus x = e2fd (i2e 0 -~ fd2e x);;

let max_cstr xs x =
  min_cstr (Array.map ~f:minus xs) (minus x);;

let max xs =
  minus (min (Array.map ~f:minus xs));;

open Printf

let domain_of var = 
  match Fd.value var with
    Val x -> Fcl_domain.create [x]
  | Unk d -> Attr.dom d;;

class element index array value =
  let n = Array.length array in
  let value_exp = fd2e value in

  let bound_index solved i =
    Fcl_stak.set solved true;
    assert(i >= 0 && i < n);
    Fcl_debug.call 'e' (fun s -> fprintf s "Element: bound_index=%d\n" i);
    Fcl_cstr.post (value_exp =~ fd2e array.(i));
    Fcl_debug.call 'e' (fun s -> fprintf s "value=%a\n" Fd.fprint value) in

  object
  inherit Fcl_cstr.t Fcl_cstr.normal

  method name = "Element.cstr"
  method delay x =
    delay [Attr.on_refine] index x;
    delay [Attr.on_refine] value x;
    Array.iter ~f:(fun v -> delay [Attr.on_refine] v x) array
  method init =
    match Fd.value index with
      Val i ->
	Fcl_stak.set solved true;
	if i >= 0 && i < n then begin
	  bound_index solved i
	end else Fcl_stak.fail "Element.get#init: index out of bound"
    | Unk index_ ->
	Fd.refine index (Fcl_domain.intersection (Fcl_domain.interval 0 (n-1)) (Attr.dom index_))
  method update =
    Fcl_debug.call 'e' (fun s -> fprintf s "[|%a|].(%a) = %a\n" (fun s -> Array.iter ~f:(fun v -> fprintf s "%a " Fd.fprint v)) array Fd.fprint index Fd.fprint value);
    let dom_value = domain_of value in
    let index_to_keep = ref []
    and new_dom_value = ref Fcl_domain.empty in
    match Fd.value index with
      Val i ->
	bound_index solved i
    | Unk index_ ->
	Attr.iter index_
	  (fun i ->
	    let inter = Fcl_domain.intersection (domain_of array.(i)) dom_value in
	    if not (Fcl_domain.is_empty inter) then begin
	      index_to_keep := i :: !index_to_keep;
	      new_dom_value := Fcl_domain.union inter !new_dom_value
	    end);
	Fd.refine index (Fcl_domain.create !index_to_keep);
	match Fd.value index with
	  Val i ->
	    bound_index solved i
    	| Unk _index_ ->
	    match Fd.value value with
	      Val _ -> Fcl_debug.call 'e' (fun s -> fprintf s "After: [|%a|].(%a) = %a\n" (fun s -> Array.iter ~f:(fun v -> fprintf s "%a " Fd.fprint v)) array Fd.fprint index Fd.fprint value) (* Something more to do ? *)
	    | Unk _ ->
		Fd.refine value !new_dom_value;
		Fcl_debug.call 'e' (fun s -> fprintf s "After: [|%a|].(%a) = %a\n" (fun s -> Array.iter ~f:(fun v -> fprintf s "%a " Fd.fprint v)) array Fd.fprint index Fd.fprint value);
end

let get array index =
  if Array.length array = 0 then invalid_arg "FdArray.get";
  let i = Fd.min index in
  if i = Fd.max index then begin (* Index connu *)
    if i >= 0 && i < Array.length array then array.(i) else Fcl_stak.fail "Element.get: index out of bound"
  end else
    let mi, ma =
      Array.fold_left
      ~f:(fun (mi, ma) e ->
	    (Pervasives.min mi (Fd.min e), Pervasives.max ma (Fd.max e)))
      ~init:(max_int, min_int) array in
    let value = Fd.create (Fcl_domain.interval mi ma) in
    Fcl_cstr.post (new element index array value);
    value;;
  

let get_cstr array index value =
  if Array.length array = 0 then invalid_arg "FdArray.get_cstr";
  new element index array value

