(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, ENAC/CENA          *)
(*                                                                     *)
(***********************************************************************)
(* $Id: fcl_goals.mli,v 1.29 2001/01/29 14:22:48 brisset Exp $ *)

(* Module [Goals]: building and solving goals *)

(*** Overview *)
(*q This module provides functions and operators to build goals that will
   control the search, i.e. mainly choose and instantiate variables.
*)

(*** Goals type, creation, ... *)
type t
  (* The type of goals. *)
val create : ?name:string -> (unit -> unit) -> t
  (* [create (?name:"create") f] returns a goal calling function [f].
     [f] must take [()] as argument and return [()]. [name] default
     value is ["create"]. *)
val make : ?name:string -> ('a -> t option) -> 'a -> t
  (* [make (?name:"make") f a] returns a goal calling [f a].
     [f] may return a goal (embedded in an option type). [name]
     default value is ["make"]. *)
val make_rec : ?name:string -> (t -> t option) -> t
  (* [make_rec (?name:"make_rec") f] returns a goal calling [f]. [f]
     takes itself as argument and may return a goal (embedded in an option type).
     Useful to write recursive goals. [name] default value is ["make_rec"]. *)
val fail : t
  (* Failure. Neutral element for the disjunction over goals. Could be 
     implemented as [create (fun () -> Stak.fail "fail")]. *)
val success : t
  (* Success. Neutral element for the conjunction over goals. Could be 
     implemented as [create (fun () -> ())]. *)
val (&&~) : t -> t -> t
  (* Conjunction over goals. *)
val (||~) : t -> t -> t
  (* Disjunction over goals. *)
val once : t -> t
  (* [once g] cuts choice points left on goal [g]. *)

val solve : ?control:(int -> unit) -> t -> bool
  (* [solve ?control g] solves the goal [g] and returns a success ([true]) or a
     failure ([false]). The execution can be controlled
     precisely with the [control] argument whose single argument is the number
     of bactracks since the beginning of the search. This function is called
     after every local failure:

   - it can raise [Stak.Fail] to force a failure of the search in the
     current branch (i.e. backtrack);

   - it can raise any other user exception to stop the search process;

   - it must return [unit] to continue the search; this is the default
     behavior. *)
(*{*)
val reset : unit -> unit
  (* _Undocumented_
     Resets the OR stack (it is not done by solve). *)
val save : t list -> Fcl_stak.level
  (* _Undocumented_
     Push a goal list on the OR stack. For advanced users. *)
(*}*)

val unify : Fcl_var.Fd.t -> int -> t
(* [unify var x] instantiates variable [var] to [x] *)

val indomain : Fcl_var.Fd.t -> t
(* Non-deterministic instanciation of a variable, by labeling its domain
   (in increasing order). *)

val instantiate : (Fcl_domain.t -> int) -> Fcl_var.Fd.t -> t
(* [instantiate choose var] Non-deterministic instanciation of a variable,
   by labeling its domain using the value returned by the [choose] function. *)

val dichotomic : Fcl_var.Fd.t -> t
(* Non-deterministic instanciation of a variable, by dichotomic recursive
   exploration of its domain. *)

val forto : int -> int -> (int -> t) -> t
  (* [forto min max g] conjunctive iteration of goal [g] on increasing
     integers from [min] to [max]. *)
val fordownto : int -> int -> (int -> t) -> t
  (* [fordownto min max g] conjunctive iteration of goal [g] on
     decreasing integers from [max] to [min]. Equivalent to
     [forto min max (fun i -> g (max - i + min))]*)


(*** Goal operation on array of variables. *)
module GlArray : sig
  val iter : ('a -> t) -> 'a array -> t
  (* [iter g a] returns the conjunctive iteration of goal [g] on array [a]. [iter g [|x1;x2;...;xn|]] is [g x1 &&~ g x2 &&~ ... &&~ g xn] *)
  val iteri : (int -> 'a -> t) -> 'a array -> t
  (* [iter g a] returns the conjunctive iteration of goal [g] on
     integer step of iteration [i] and array [a]. [iteri g [|x1;x2;...;xn|]] is [g 0 x1 &&~ g 1 x2 &&~ ... &&~ g (n-1) xn] *)
(*{*)
  val iter2 : ('a -> 'b -> t) -> 'a array -> 'b array -> t
(* Same as [iter] but on two arrays. *)
(*}*)
  val labeling : Fcl_var.Fd.t array -> t
  (* Equivalent to [iter indomain]. *)
  val iter_h : ('a array -> int) -> ('a -> t) -> 'a array -> t
  (* [iter_aray_h h g a] returns the conjunctive iteration
     of goal [g] where order is computed by the heuristic [h] which
     must raise [Not_found] to terminate, i.e. at each iteration
     [g] is applied to [a.(h a)]. *)
  val iter_hi : ('a array -> int) -> (int -> 'a -> t) -> 'a array -> t
  (* [iter_aray_hi h g a] returns the conjunctive iteration
     of goal [g] where order is computed by the heuristic [h] which
     must raise [Not_found] to terminate, i.e. at the [i]th iteration
     [g] is applied to [a.(h i a)]. *)
  val choose_index : (Fcl_var.Attr.t -> Fcl_var.Attr.t -> bool) -> Fcl_var.Fd.t array -> int
  (* [choose_index order fds] returns the index of the best (minimun)
     free (not instantiated) variable in the array [fds] for the criterion
     [order]. Raises [Not_found] if all variables are bound (instantiated). *)
  val not_instantiated_fd : Fcl_var.Fd.t array -> int
  (* [not_instantiated_fd fds] returns the index of one element in [fds]
     which is not instantiated. Raises [Not_found] if all variables in array
     [fds] are bound. *)
end

(*** Goal operations on list of variables. *)
module GlList : sig
  val iter : ('a -> t) -> 'a list -> t
  (* [iter g [x1;x2;...;xn]] is [g x1 &&~ g x2 &&~ ... &&~ g xn], i.e.
     returns the conjunctive iteration of goal [g] on list [a]. *)
  val labeling : Fcl_var.Fd.t list -> t
    (* Equivalent to [iter indomain]. *)
(*{*)
  val iter_disj : ('a -> t) -> 'a list -> t
  (* [iter_disj g [x1;x2;...;xn]] is [g x1 ||~ g x2 ||~ ... ||~ g xn], i.e.
     return the disjunctive iteration of goal [g] on list [a]. *)
  val member : Fcl_var.Fd.t -> int list -> t
(* [member v l] returns the disjunctive iteration on the instantiation of
   the variable [v] to the values in the integer list [l]. Defined by
   [fun v -> iter_disj (fun x -> create (fun () -> Fd.unify v x))] *)
(*}*)
  val iter_h : ('a list -> 'a*('a list)) -> ('a -> t) -> 'a list -> t
  (* [iter_h h g l] returns the conjunctive iteration of goal [g] where
     order is computed by the heuristic [h], i.e. at each iteration
     [h l] returns the couple [(v,vs)] of the variable (or data) which
     will be passed as the argument of goal [g] (thus executing [g v]) and
     the list [vs] on which the iteration will go on. *)
end

  

