(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, ENAC/CENA          *)
(*                                                                     *)
(***********************************************************************)
(* $Id: fcl_var.ml,v 1.14 2001/01/24 13:59:01 barnier Exp $ *)

(* Un objet avec des contraintes qui y sont attaches *)
class cstrd () = (* Parameter is needed because we do not want cstrs as a class
		    variable *)
  let cstrs = Fcl_stak.ref [] in
 object
  method delayed = Fcl_stak.get cstrs

  method delay (cstr : Fcl_cstr.t) =
    let current = Fcl_stak.get cstrs
    and cstr_id = cstr#id in
    if not cstr#is_solved && not (List.exists ~f:(fun c -> c#id = cstr_id) current)
    then Fcl_stak.set cstrs (cstr :: current)

  method wake =
    List.iter ~f:Fcl_cstr.wake (Fcl_stak.get cstrs)
    
  method constraints_number =
    List.fold_left ~f:(fun r x -> if x#is_solved then r else 1+r) ~init:0 (Fcl_stak.get cstrs)

  method constraints_id =
    List.map ~f:(fun c -> c#id) (Fcl_stak.get cstrs)
end
      
module Attr = struct
  type t = {
      dom : Fcl_domain.t Fcl_stak.ref;
      on_refine : cstrd; 
      on_min : cstrd; 
      on_max : cstrd; 
      on_subst : cstrd;
      id : int
    }

  let create domain =
    {
      dom = Fcl_stak.ref domain;
      on_refine = new cstrd (); 
      on_min = new cstrd (); 
      on_max = new cstrd (); 
      on_subst = new cstrd ();
      id = Fcl_misc.gen_int ()
    } 

  type event = t -> cstrd

  let dom a = Fcl_stak.get a.dom
  let on_refine a = a.on_refine
  let on_subst a = a.on_subst
  let on_min a = a.on_min
  let on_max a = a.on_max
  let fprint c a = Printf.fprintf c "%d{%a}" a.id Fcl_domain.fprint (Fcl_stak.get a.dom)
  let size a = Fcl_domain.size (Fcl_stak.get a.dom)
  let min a = Fcl_domain.min (Fcl_stak.get a.dom)
  let max a = Fcl_domain.max (Fcl_stak.get a.dom)
  let values a = Fcl_domain.values (Fcl_stak.get a.dom)
  let iter a f = Fcl_domain.iter f (Fcl_stak.get a.dom)
  let member a y = Fcl_domain.member y (Fcl_stak.get a.dom)
  let id a = a.id

(* All constraints are suspended on on_subst *)
  let constraints_number a =
    List.length (Fcl_misc.remove_duplicates (on_subst a)#constraints_id)
end


type concrete_fd = Unk of Attr.t | Val of int

module Fd = struct
  type t = concrete_fd Fcl_stak.ref
  let int i = Fcl_stak.ref (Val i)
  let create domain =
    Fcl_stak.ref
      (match Fcl_domain.size domain with
	   0 -> Fcl_stak.fail "Var.Fd.create: empty initial domain"
	 | 1 -> Val (Fcl_domain.min domain)
	 | _ -> Unk (Attr.create domain))

  let interval min max = create (Fcl_domain.interval min max)

  let array n min max =
    let dom = Fcl_domain.interval min max in
    Array.init n ~f:(fun _ -> create dom)

  let subst v new_v =
    match Fcl_stak.get v with
      Unk a ->
    	Fcl_debug.call 't' (fun s -> Printf.fprintf s "subst _%d%a with %d\n" a.Attr.id Attr.fprint a new_v);
      	if Fcl_domain.member new_v (Attr.dom a)
      	then begin
	  Fcl_stak.set v (Val new_v);
	  (Attr.on_subst a)#wake; (Attr.on_refine a)#wake;
	  (Attr.on_min a)#wake; (Attr.on_max a)#wake;
	  Fcl_cstr.wake_all ()
      	end
      	else Fcl_stak.fail "Var.Fd.subst"
    | Val _ ->
	Fcl_debug.fatal_error "Fd.subst: bound variable (use Fd.unify on possible bound variable)"
    
  let value (v : t) = Fcl_stak.get v

  let refine v new_a =
    match value v with
      Unk a ->
      	let vala = Attr.dom a in
	Fcl_debug.call 't' (fun s -> Printf.fprintf s "refine _%d%a with %a\n" a.Attr.id Fcl_domain.fprint vala Fcl_domain.fprint new_a);
      	assert (Fcl_domain.included new_a vala);
      	let new_size = Fcl_domain.size new_a in
      	if new_size = 0
      	then Fcl_stak.fail "Var.Fd.refine"
      	else if new_size = Fcl_domain.size vala
      	then ()
      	else if new_size = 1
      	then subst v (Fcl_domain.min new_a)
      	else begin
      	  Fcl_stak.set a.Attr.dom new_a;
      	  let min = Fcl_domain.min vala and max = Fcl_domain.max vala in
      	  a.Attr.on_refine#wake;
      	  if Fcl_domain.min new_a > min then a.Attr.on_min#wake;
      	  if Fcl_domain.max new_a < max then a.Attr.on_max#wake;
	  Fcl_cstr.wake_all ()
      	end
    | Val _ ->
	Fcl_debug.fatal_error "Fd.refine: bound variable"

  let unify (v : t) new_v =
    match Fcl_stak.get v with
      Val v -> if not (v = new_v) then Fcl_stak.fail "Var.Fd.unify"
    | Unk _a -> subst v new_v
	  
  let fprint c v =
    match Fcl_stak.get v with
      Val t -> Printf.fprintf c "%d" t
    | Unk a -> Printf.fprintf c "_%a" Attr.fprint a

  let fprint_array c vs =
    let n = Array.length vs in
    Printf.fprintf c "[|";
    for i = 0 to n - 2 do
      Printf.fprintf c "%a; " fprint vs.(i)
    done;
    if n = 0 then Printf.fprintf c "|]" else
    Printf.fprintf c "%a|]" fprint vs.(n-1)

  let size v =
    match Fcl_stak.get v with
      Val _ -> 1
    | Unk a -> Attr.size a
	  
  let min v =
    match Fcl_stak.get v with
      Val t -> t
    | Unk a -> Fcl_domain.min (Fcl_stak.get a.Attr.dom)
	  
  let max v =
    match Fcl_stak.get v with
      Val t -> t
    | Unk a -> Attr.max a
	  
  let values v =
    match Fcl_stak.get v with
      Val t -> [t]
    | Unk a -> Attr.values a
	  
  let is_var v =
    match Fcl_stak.get v with
      Val _x -> false
    | Unk _a -> true
	  
  let id v =
    match Fcl_stak.get v with
      Val _x -> Fcl_debug.fatal_error "Fd.id: bound variable"
    | Unk a -> a.Attr.id

  let iter v f =
    match Fcl_stak.get v with
      Val x -> f x
    | Unk a -> Attr.iter a f

  let member v x =
    match Fcl_stak.get v with
      Val v -> v = x
    | Unk a -> Attr.member a x

  let compare v1 v2 =
    match Fcl_stak.get v1, Fcl_stak.get v2 with
      (Val n1, Val n2) -> compare n1 n2
    | (Val _, Unk _) -> -1
    | (Unk _, Val _) -> 1
    | (Unk a1, Unk a2) -> compare (Attr.id a1) (Attr.id a2)

  let equal v1 v2 = compare v1 v2 = 0
end
	  

let delay events x c =
  match (Fd.value x) with
    Val _ -> ()
  | Unk x_ -> List.iter ~f:(fun event -> (event x_)#delay c) events
