(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, ENAC/CENA          *)
(*                                                                     *)
(***********************************************************************)
(* $Id: fcl_var.mli,v 1.8 2001/01/29 14:46:30 brisset Exp $ *)

(* Module [Var]: constrained, attributed, finite domain variables *)

(*{*)
(* A constrained object *)
type cstrd
(*}*)

(*** The attribute of a domain variable *)
module Attr : sig
  type t
(* Type of attributes. *)
  val dom : t -> Fcl_domain.t
(* [dom a] returns the integer domain of an attribute. *)
  type event
(* Type of events (modifications on variables) on which to suspend
   a constraint. *)
  val on_refine : event
(* Event occuring when a variable is changed, i.e. its domain modified. *)
  val on_subst : event
(* Event occuring when a variable is instantiated. *)
  val on_min : event
(* Event occuring when the lower bound of a variable increases. *)
  val on_max : event
(* Event occuring when the upper bound of a variable decreases. *)
  val fprint : out_channel -> t -> unit
(* [fprint chan a] prints attribute [a] on channel [chan]. *)
  val size : t -> int
(* [size a] returns the number of integer values in the domain associated with
   [a] (i.e. [dom a]). *)
  val min : t -> int
(* [min a] returns the lower bound of [a]. *)
  val max : t -> int
(* [max a] returns the upper bound of [a]. *)
  val values : t -> int list
(* [values a] returns the list of all integers in [dom a]. *)
  val iter : t -> (int -> unit) -> unit
(* [iter a f] iters f on each integer in [dom a]. *)
  val member : t -> int -> bool
(* [member a n] tests if [n] belongs to [dom a]. *)
  val id : t -> int
(* [id a] returns a unique integer identifying the attribute [a]. *)
  val constraints_number : t -> int
(* [constraints_number a] returns the number of different constraints
   attached to [a]. *)
end

(*** A domain variable *)

type concrete_fd = Unk of Attr.t | Val of int
(* Type of the value of a Fd variable. *)

module Fd : sig
(* Finite domain variable *)
  type t
(* Type of Fd variable. *)
  val create : Fcl_domain.t -> t
(* [create d] returns a new variable with domain [d]. *)
  val interval : int -> int -> t
(* [interval inf sup] returns a new variable with domain [inf..sup]. *)
  val array : int -> int -> int -> t array
(* [array n inf sup] returns an array of [n] new variables with domain
   [inf..sup]. *)
  val int : int -> t
(* [int n] returns a new variable instantiated to integer value [n]. *)
  val is_var : t -> bool
(* [is_var v] tests if [v] is not instantiated. *)
  val value : t -> concrete_fd
(* [value v] returns [Val n] if [v] is instantiated to [n], [Unk a] otherwise
   where [a] is the attribute of [v]. Should always be used in a matching:
   [match value v with Val n -> ... | Unk a -> ...]. *)
  val subst : t -> int -> unit
(* [subst v n] instantiates variable [v] with integer value [n]. Raises
   [Fcl_stak.Fail] in case of failure. Must be called only on unbound
   (not instantiated) variable, otherwise a [Failure] exception is raised. *)
  val refine : t -> Fcl_domain.t -> unit
(* [refine v d] reduces the domain of [v] with domain [d]. [d] must be
   included in the domain of [v], otherwise the behaviour is
   unspecified (corrupted system or exception raised).
   Must be called only on non ground variable ([Failure] exception
   otherwise). *)
  val unify : t -> int -> unit
(* Idem as subst but can be called on instantiated variable. *)
  val fprint : out_channel -> t -> unit
(* [fprint chan v] prints variable [v] on channel [chan]. *)
  val fprint_array : out_channel -> t array -> unit
(* [fprint_array chan vs] prints array of variables [vs] on channel [chan]. *)
  val size : t -> int
(* [size v] returns the number of integer values in the domain of [v]
   ([1] if [v] is instantiated. *)
  val min : t -> int
(* [min v] returns the lower bound of [v]. *)
  val max : t -> int
(* [max v] returns the upper bound of [v]. *)
  val values : t -> int list
(* [values v] returns the list of all integers in the domain of [v]. If
   [v] is instantiated to [n], returns the singleton list containing [n]. *)
  val iter : t -> (int -> unit) -> unit
(* [iter v f] iters f on each integer in the domain of [v]. *)
  val member : t -> int -> bool
(* [member v n] tests if [n] belongs to the domain of [v]. *)
  val id : t -> int
(* [id v] returns a unique integer identifying the attribute associated
   with [v]. Must be called only on non ground variable, raise Failure
   otherwise. *)
  val compare : t -> t -> int
  (* Compares two variables. Values (bound variables) are smaller than
     unknowns (unbound variables). Unknowns are sorted according to
     their attribute [id]. *)
  val equal : t -> t -> bool
  (* Tests if two variables are equal with respect to [compare]. *)
end

val delay : Attr.event list -> Fd.t -> Fcl_cstr.t -> unit
(* [delay event_list v c] suspends constraint [c] on all the events in
   [event_list] occuring on [v]. No effect on instantiated variables. *)
